use crate::Asset;
use serde::Serialize;

pub type AmountCents = i128;

/// Amount
///
/// The cents are stored in their lowest denomination, or their "cents".  For
/// instance a Bitcoin would be represented with a `precision` of 8, or dollars
/// in a 2 or 3.
///
/// For instance, if the dollar Asset has a `precision` of `2`, an `cents` of
/// 1255 will represent `12.55 USD`.
///
/// Amount will abstract all the math operations that can be performed with the
/// amounts, guarding that the `Asset` are the same before performing any
/// operation.
///
///
/// The `cents` and `Asset.id` must be used to store amounts in the storge
/// layer. Float or string representations should be used to display
#[derive(Copy, Clone, Debug, Eq, PartialEq, Serialize)]
pub struct Amount {
    asset: Asset,
    cents: AmountCents,
}

impl Amount {
    pub fn new(asset: Asset, cents: AmountCents) -> Self {
        Self { asset, cents }
    }

    #[inline]
    pub fn asset(&self) -> &Asset {
        &self.asset
    }

    #[inline]
    pub fn cents(&self) -> AmountCents {
        self.cents
    }
}
