use crate::{
    amount::AmountCents, storage, AccountId, Amount, AssetManager, Batch, Error, PaymentId,
    Storage, Transaction,
};
use futures::TryStreamExt;
use std::{collections::HashMap, sync::Arc};

#[derive(Clone, Debug)]
pub struct Ledger<B, I, S>
where
    B: Batch,
    I: TryStreamExt<Ok = (PaymentId, AmountCents), Error = storage::Error> + Unpin,
    S: Storage<B, I>,
{
    storage: Arc<S>,
    asset_manager: Arc<AssetManager>,
    _phantom_batch: std::marker::PhantomData<B>,
    _phantom_iterator: std::marker::PhantomData<I>,
}

impl<B, I, S> Ledger<B, I, S>
where
    B: Batch,
    I: TryStreamExt<Ok = (PaymentId, AmountCents), Error = storage::Error> + Unpin,
    S: Storage<B, I>,
{
    pub fn new(storage: S, asset_manager: AssetManager) -> Self {
        Self {
            storage: Arc::new(storage),
            asset_manager: Arc::new(asset_manager),
            _phantom_batch: std::marker::PhantomData,
            _phantom_iterator: std::marker::PhantomData,
        }
    }

    pub async fn get_balance(&self, account: AccountId) -> Result<Vec<Amount>, Error> {
        self.storage
            .get_balance(account)
            .await?
            .into_iter()
            .map(|(asset, cents)| self.asset_manager.amount(asset, cents))
            .collect::<Result<Vec<_>, _>>()
    }

    async fn normalize_payments_to_spend(
        &self,
        payments: Vec<(AccountId, Amount)>,
    ) -> Result<Vec<(AccountId, Amount)>, Error> {
        let mut to_spend = HashMap::new();
        let mut balances = HashMap::new();
        for (account_id, amount) in payments.into_iter() {
            if balances.get(&account_id).is_none() {
                balances.insert(account_id, self.get_balance(account_id).await?);
            }
            let id = (account_id, *amount.asset());
            if let Some(value) = to_spend.get_mut(&id) {
                *value += amount.cents();
            } else {
                to_spend.insert(id, amount.cents());
            }
        }

        for ((account, asset), mut to_spend_cents) in to_spend.into_iter() {
            let mut payment_ids: Vec<PaymentId> = vec![];
            let mut iterator = self.storage.get_unspend_payments(account, asset.id).await?;
            while let Some((payment_id, cents)) = iterator.try_next().await? {
                payment_ids.push(payment_id);
                to_spend_cents -= cents;
                if to_spend_cents < 0 {
                    // We have enough payment_id to to cover the to_spend_cents
                    // Any negative number will be the return to the account
                    break;
                }
            }
        }

        /*
               to_spend
                   .into_iter()
                   .map(|((account_id, asset), cents)| (account_id, Amount::new(asset, cents)))
                   .collect::<Vec<_>>()
                   .into_iter()
                   .map(|(id, to_debit)| {
                       balances
                           .get(&id)
                           .ok_or_else(|| Error::InsufficientBalance(id, to_debit.asset().id))
                           .map(|balances| {
                               if let Some(available) = balances
                                   .iter()
                                   .find(|asset| asset.asset() == to_debit.asset())
                                   .map(|amount| amount.cents())
                               {
                                   if to_debit.cents() > available {
                                       Err(Error::InsufficientBalance(id, to_debit.asset().id))
                                   } else {
                                       Ok((id, to_debit))
                                   }
                               } else {
                                   Err(Error::InsufficientBalance(id, to_debit.asset().id))
                               }
                           })
                   })
                   .collect::<Result<Result<Vec<_>, _>, _>>()?
        */
        todo!()
    }

    pub async fn new_transaction<'a>(
        &self,
        from: Vec<(AccountId, Amount)>,
        to: Vec<(AccountId, Amount)>,
    ) -> Result<Transaction<'a, B, I, S>, Error> {
        let from = self.normalize_payments_to_spend(from).await?;
        todo!()
    }
}
