use std::marker::PhantomData;

use crate::{
    storage::{self, Error},
    AccountId, Payment, PaymentId,
};
use sqlx::{Sqlite, Transaction as SqlxTransaction};

pub struct Batch<'a> {
    inner: SqlxTransaction<'a, Sqlite>,
    x: PhantomData<&'a ()>,
}

#[async_trait::async_trait]
impl<'a> storage::Batch for Batch<'a> {
    async fn spend_payment(
        &mut self,
        payment_id: PaymentId,
        transaction_id: AccountId,
    ) -> Result<(), Error> {
        let result = sqlx::query(
            r#"
                UPDATE payments SET spent_by = ?
                WHERE transaction_id = ? AND position_id = ? AND spent_by IS NULL
            "#,
        )
        .bind(transaction_id.to_string())
        .bind(payment_id.transaction.to_string())
        .bind(payment_id.position.to_string())
        .execute(&mut *self.inner)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;
        if result.rows_affected() == 1 {
            Ok(())
        } else {
            Err(Error::NoUpdate)
        }
    }

    async fn rollback(self) -> Result<(), Error> {
        self.inner
            .rollback()
            .await
            .map_err(|e| Error::Storage(e.to_string()))
    }

    async fn commit(self) -> Result<(), Error> {
        self.inner
            .commit()
            .await
            .map_err(|e| Error::Storage(e.to_string()))
    }

    async fn store_new_payments(&mut self, payments: &[Payment]) -> Result<(), Error> {
        for payment in payments.iter() {
            sqlx::query(
                r#"
                INSERT INTO payments(transaction_id, position_id, to, amount, asset, spent_by)
                VALUES (?, ?, ?, ?, ?, ?)
            "#,
            )
            .bind(payment.id.transaction.to_string())
            .bind(payment.id.position.to_string())
            .bind(payment.to.to_string())
            .bind(payment.amount.cents().to_string())
            .bind(payment.amount.asset().id)
            .bind(payment.spent_by.map(|id| id.to_string()))
            .execute(&mut *self.inner)
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;
        }
        Ok(())
    }

    async fn store_transaction(&mut self) -> Result<(), Error> {
        todo!()
    }
}

#[cfg(test)]
mod test {}
