use crate::{
    amount::AmountCents, storage, AccountId, Asset, Batch, Payment, PaymentId, Status, Storage,
};
use futures::{future::join_all, TryStreamExt};
use sha2::{Digest, Sha256};
use std::collections::HashMap;

mod error;

pub use error::Error;

/// Transactions
///
/// Transactions are the core components of the ledger. The transactions are a
/// list of unspent payments that are about to be spend, to create a new set of
/// Payments, that can be spend in the future. This model is heavily inspired in
/// Bitcoin's UTXO model. The terms in this context are payments, spend and
/// create instead of unspent transactions, input and output.
///
/// This simple architecture allows to track accounts pretty efficiently,
/// because all that matters are unspent payments owned by a given account.
/// Every spent payment is stored for historical reasons but it is not relevant
/// for any calculations regarding available funds.
///
/// The transaction has a few rules, for instance the sum of spend Payments
/// should be the same as create Payments, for each easy. There is no 'fee'
/// concept, so any mismatch in any direction will error the constructor.
///
/// Transactions are immutable after they are finalized, and the payments can
/// only be re-usable if the transaction failed or was cancelled. Once the
/// transaction settles the spent payments are forever spent. Any rollback
/// should be a new transaction, initiated by a higher layer.
///
/// The spent payments are unavailable until the transaction is finalized,
/// either as settled, cancelled or failed. A higher layer should split any
/// available payment to be spend into a new transaction, and then finalize the
/// transaction, and reserve only the exact amount to be spent, otherwise
/// unrelated funds will be held unspendable until the transaction is finalized.
#[derive(Debug, Clone)]
pub struct Transaction<'a, B, I, S>
where
    B: Batch,
    I: TryStreamExt<Ok = (PaymentId, AmountCents), Error = storage::Error> + Unpin,
    S: Storage<B, I>,
{
    storage: &'a S,
    id: Option<AccountId>,
    spend: Vec<Payment>,
    create: Vec<Payment>,
    status: Status,
    _phantom_batch: std::marker::PhantomData<B>,
    _phantom_iterator: std::marker::PhantomData<I>,
}

impl<'a, B, I, S> Transaction<'a, B, I, S>
where
    B: Batch,
    I: TryStreamExt<Ok = (PaymentId, AmountCents), Error = storage::Error> + Unpin,
    S: Storage<B, I>,
{
    pub async fn new(
        storage: &'a S,
        input_ids: Vec<PaymentId>,
    ) -> Result<Transaction<'a, B, I, S>, Error> {
        let futures = input_ids
            .into_iter()
            .map(|id| storage.get_payment(id))
            .collect::<Vec<_>>();
        let spend = join_all(futures)
            .await
            .into_iter()
            .collect::<Result<Vec<Payment>, _>>()?;
        for (i, input) in spend.iter().enumerate() {
            if input.spent_by.is_some() {
                return Err(Error::SpentPayment(i));
            }
        }

        Ok(Self {
            storage,
            id: None,
            spend,
            status: Status::default(),
            create: Vec::new(),
            _phantom_batch: std::marker::PhantomData,
            _phantom_iterator: std::marker::PhantomData,
        })
    }

    pub fn change_status(&mut self, new_status: Status) -> Result<(), Error> {
        if self.status.can_transition_to(&new_status) {
            self.status = new_status;
            Ok(())
        } else {
            Err(Error::StatusTransitionNotAllowed(
                self.status.clone(),
                new_status,
            ))
        }
    }

    fn validate(&mut self) -> Result<(), Error> {
        let mut sha256 = Sha256::new();
        let mut debit = HashMap::<Asset, AmountCents>::new();
        let mut credit = HashMap::<Asset, AmountCents>::new();

        for (i, input) in self.spend.iter().enumerate() {
            if input.spent_by.is_some() {
                return Err(Error::SpentPayment(i));
            }
            if let Some(value) = debit.get_mut(input.amount.asset()) {
                *value = input
                    .amount
                    .cents()
                    .checked_add(*value)
                    .ok_or(Error::Overflow)?;
            } else {
                debit.insert(*input.amount.asset(), input.amount.cents());
            }
            let bytes: Vec<u8> = bincode::serialize(&input).unwrap();
            sha256.update(&bytes);
        }

        for (i, output) in self.create.iter().enumerate() {
            if output.spent_by.is_some() {
                return Err(Error::SpentPayment(i));
            }
            if let Some(value) = credit.get_mut(output.amount.asset()) {
                *value = output
                    .amount
                    .cents()
                    .checked_add(*value)
                    .ok_or(Error::Overflow)?;
            } else {
                credit.insert(*output.amount.asset(), output.amount.cents());
            }
            let bytes: Vec<u8> = bincode::serialize(&output).unwrap();
            sha256.update(&bytes);
        }

        for (asset, credit_amount) in credit.into_iter() {
            if let Some(debit_amount) = debit.remove(&asset) {
                if debit_amount != credit_amount {
                    return Err(Error::InvalidAmount(asset, debit_amount, credit_amount));
                }
            } else {
                return Err(Error::MissingSpendingAsset(asset));
            }
        }

        if let Some((asset, _)) = debit.into_iter().next() {
            return Err(Error::MissingPaymentAsset(asset));
        }

        self.id = Some(AccountId(sha256.finalize().into()));
        Ok(())
    }

    pub fn spent(&self) -> &[Payment] {
        &self.spend
    }

    pub fn created(&self) -> &[Payment] {
        &self.create
    }

    pub async fn persist(mut self) -> Result<Transaction<'a, B, I, S>, Error> {
        if self.status.is_finalized() {
            return Err(Error::TransactionUpdatesNotAllowed);
        }
        self.validate()?;
        let mut batch = self.storage.begin().await?;
        batch.store_new_payments(&self.create).await?;
        let id = self.id.as_ref().unwrap();
        for input in self.spend.iter_mut() {
            batch.spend_payment(input.id, *id).await?;
            input.spent_by = Some(*id);
        }
        batch.commit().await?;
        Ok(self)
    }
}
